import databento as dbn
import os
from datetime import datetime, timedelta, timezone
from dateutil.parser import isoparse
from typing import List
import polars as pl


class DBNClient:
    """
    Databento Client Wrapper"""

    def __init__(self):

        if "DATABENTO_API_KEY" not in os.environ:
            raise ValueError("Missing Databento API key")

        self._client = dbn.Historical()
        print("Databento client initialized.")

    def _download_data(
        self,
        dataset: str,
        schema: str,
        symbol: str,
        start: str,
        end: str,
    ):
        """
        Download data from Databento
        """

        data = self._client.timeseries.get_range(
            dataset=dataset,
            schema=schema,
            symbols=symbol,
            start=start,
            end=end,
        ).to_df()
        return data

    def _get_dataset_range(
        self,
        dataset: str,
    ) -> tuple[datetime, datetime]:
        """
        Get the range of a dataset
        """
        range = self._client.metadata.get_dataset_range(dataset)
        start_str = range["start"]
        end_str = range["end"]
        start = isoparse(start_str)
        end = isoparse(end_str)

        if (
            end.hour == 4
            and end.minute == 0
            and end.second == 0
            and end.microsecond == 0
        ):
            prev = end - timedelta(days=1)
            end = datetime(
                prev.year,
                prev.month,
                prev.day,
                23,
                58,
                tzinfo=None,
            )

        return start, end

    def _get_columns(
        self,
        dataset: str,
        schema: str | List[str],
    ) -> dict[list[str]]:
        """
        Get the columns in the schema from Databento and prepare them for Database
        """
        if isinstance(schema, str):
            schema = [schema]

        col_dict = {}

        # validate dataset and schema
        for s in schema:
            cols = self._client.metadata.list_fields(s, "csv")

            for col in cols:
                if col["name"] in ("ts_event", "ts_recv"):
                    col["type"] = "timestamp"

            # symbol always appears at the end
            cols.append({"name": "symbol", "type": "string"})
            col_dict[s] = cols

        return col_dict

    def _validate_schema_and_dataset(self, dataset: str, schema: list[str]) -> None:
        """
        Validate the dataset and schema in Databento
        """
        if not self._validate_dataset(dataset):
            raise ValueError(f"Dataset {dataset} not found in Databento.")
        for s in schema:
            if not self._validate_schema(dataset, s):
                raise ValueError(f"Schema {s} not found in Databento.")

    def _validate_dataset(self, dataset: str) -> bool:
        if dataset not in self._client.metadata.list_datasets():
            print(f"Dataset {dataset} not found in Databento.")
            return False
        return True

    def _validate_schema(self, dataset: str, schema: str) -> bool:
        if schema not in self._client.metadata.list_schemas(dataset):
            print(f"Schema {schema} not found in Databento.")
            return False
        return True

    def _calculate_cost(
        self,
        dataset: str,
        schema: str,
        symbol: str,
        start: str,
        end: str,
    ) -> float:
        """
        Calculate the cost of downloading data from Databento
        """

        # Get the size of the data
        cost = self._client.metadata.get_cost(
            dataset=dataset,
            schema=schema,
            symbols=symbol,
            start=start,
            end=end,
        )
        return cost



from pathlib import Path
import json
import os

CONFIG_PATH = CONFIG_PATH = Path.home() / ".acedb" / "config.json"


class Config:

    host: str = None
    port: int = None
    db_name: str = None
    username: str = None
    password: str = None

    def __init__(self):
        if not CONFIG_PATH.exists():
            raise FileNotFoundError(f"Configuration file not found: {CONFIG_PATH}")
        with open(CONFIG_PATH, "r") as config_file:
            raw_config = json.load(config_file)

        self.password = raw_config.get("password")

        if "dbn_token" in raw_config:
            os.environ["DATABENTO_API_KEY"] = raw_config["dbn_token"]

        self.host = raw_config.get("host")
        self.port = raw_config.get("port")
        self.db_name = raw_config.get("db_name")
        self.username = raw_config.get("username")



from .config import Config
from typing import List, Dict, Any, Tuple
from collections import defaultdict

import pandas as pd

from .postgreclient import PostgreDBClient
from .dbnclient import DBNClient


class AceDB:

    def __init__(self):
        self._config = Config()
        self.database_client = PostgreDBClient(
            host=self._config.host,
            port=self._config.port,
            db_name=self._config.db_name,
            username=self._config.username,
            password=self._config.password,
        )
        self.databento_client = DBNClient()

    def get_data(
        self,
        dataset: str,
        schemas: List[str] | str,
        symbols: List[str] | str,
        start: str = None,
        end: str = None,
        use_databento: bool = True,
        download: bool = False,
        path: str = None,
        filetype: str = "csv",
        **kwargs,
    ):
        """
        Retrieves data from Database and Databento to find minimum Cost.

        It will check if the data is already in the database. If not, it will
        download the data from Databento and insert it into the database.
        If the data is already in the database, it will retrieve it from there.


        Parameters
        ----------
        dataset : str
            This is the name of the dataset as found on databento.
        schemas : List[str] | str
            This is a schema as a string or multiple schemas as a list of strings, as found on databento.
        symbols : List[str] | str
            This is a symbol as a string or multiple symbols as a list of strings. These are Stock Tickers.
        start : str, optional
            This is the start date of the data to be retrieved. The default is None.
        end : str, optional
            This is the end date of the data to be retrieved. The default is None.

        Returns
        -------
        dict[str, pd.DataFrame]
            A dictionary of DataFrames, where the keys are the schema names and the values are the DataFrames.
            Each DataFrame contains the data for the corresponding schema and symbol.


        Examples
        --------
        >>> acedbget_data(dataset='XNAS.ITCH',schemas='ohlcv-1h',symbols='NVDA',start='2023-01-01',end='2023-01-02')
        """
        # Makes sure schemas and symbols are lists
        schemas = [schemas] if isinstance(schemas, str) else schemas
        symbols = [symbols] if isinstance(symbols, str) else symbols

        if not isinstance(schemas, list) or not isinstance(symbols, list):
            raise ValueError("Schemas and symbols must be lists or strings.")

        # Check if the dataset is valid and exist on databento
        self.databento_client._validate_schema_and_dataset(
            schema=schemas, dataset=dataset
        )

        # Returns the columns of the schema from databento
        col_dict = self.databento_client._get_columns(
            dataset=dataset,
            schema=schemas,
        )

        # Check if the dataset is valid and exist on database and create the database if not
        self.database_client._ensure_database_schema(
            dataset=dataset, schemas=schemas, col_dict=col_dict
        )

        result = {}

        if not use_databento:
            for schema in schemas:

                result[schema] = {}

                for symbol in symbols:

                    result[schema][symbol] = self.database_client._retrieve_data(
                        dataset=dataset,
                        schema=schema,
                        symbol=symbol,
                    )

            if download:
                print("Downloading data...")
                self.database_client._download(
                    data_dict=result,
                    path=path,
                    filetype=filetype,
                )

            return result

        # Get the start and end date of the dataset
        min_start, max_end = self.databento_client._get_dataset_range(dataset=dataset)

        for schema in schemas:

            result[schema] = {}

            for symbol in symbols:
                # Find the start and end date of the schema/symbol in the database
                db_start, db_end = self.database_client._get_local_range(
                    dataset=dataset,
                    schema=schema,
                    symbol=symbol,
                )

                # Finds ranges it needs to download from databento
                download_ranges = self.database_client._find_missing_range(
                    dataset_range=(min_start, max_end),
                    database_range=(db_start, db_end),
                    query_range=(start, end),
                )

                # If there are no ranges to download, return the data from the database
                for range_start, range_end in download_ranges:
                    cost = self.databento_client._calculate_cost(
                        dataset=dataset,
                        schema=schema,
                        symbol=symbol,
                        start=range_start,
                        end=range_end,
                    )

                    # Checks if the cost is above a threshold so it doesn't download single data points
                    if cost > 1e-5:
                        if not self._ask_yn(
                            f"Download {schema}/{symbol} from {range_start} to {range_end}? Cost: {cost} [Y/n]"
                        ):
                            continue

                        print("Downloading data...")
                        data = self.databento_client._download_data(
                            dataset=dataset,
                            schema=schema,
                            symbol=symbol,
                            start=range_start,
                            end=range_end,
                        )

                        print("Inserting data into database...")
                        self.database_client._insert_database(
                            dataset=dataset,
                            schema=schema,
                            data=data,
                        )

                        print("Data downloaded and inserted into database.")

                result[schema][symbol] = self.database_client._retrieve_data(
                    dataset=dataset,
                    schema=schema,
                    symbol=symbol,
                )
        if download:
            print("Downloading data...")
            self.database_client._download(
                data_dict=result, path=path, filetype=filetype
            )
        return result

    def insert(self, dataset: str, schema: str, data: pd.DataFrame) -> None:
        """
        Insert data into the database.

        Parameters
        ----------
        dataset : str
            The name of the dataset.
        schema : str
            The name of the schema.
        data : pd.DataFrame
            The data to be inserted.

        Returns
        -------
        None
        """
        if not isinstance(data, pd.DataFrame):
            raise ValueError("Data must be a pandas DataFrame.")

        self.databento_client._validate_schema_and_dataset(
            schema=schema, dataset=dataset
        )
        col_dict = self.databento_client._get_columns(
            dataset=dataset,
            schema=schema,
        )

        assert (
            data.reset_index() == col_dict[schema]
        ), "DataFrame columns do not match schema columns."

        self.database_client._ensure_database_schema(
            dataset=dataset,
            schemas=[schema],
            col_dict={schema: col_dict},
        )

        self.database_client._insert_database(
            dataset=dataset,
            schema=schema,
            data=data,
        )
        print("Data inserted into database.")

    def get_ranges(
        self,
    ):
        result = {}

        datasets = self.database_client._get_datasets()

        datasets = [
            dataset
            for dataset in datasets
            if dataset
            not in ["public", "qrg_database_s1", "information_schema", "pg_catalog"]
        ]

        for dataset in datasets:

            schemas = self.database_client._get_schemas(dataset=dataset)
            result[dataset] = {}

            for schema in schemas:
                symbols = self.database_client._get_symbols(
                    dataset=dataset,
                    schema=schema,
                )

                result[dataset][schema] = {}
                for symbol in symbols:
                    start, end = self.database_client._get_local_range(
                        dataset=dataset,
                        schema=schema,
                        symbol=symbol,
                    )
                    result[dataset][schema][symbol] = (start, end)
        return result

    @staticmethod
    def _ask_yn(question: str) -> bool:
        """
        Ask a yes or no question and waits for an answer.
        """
        while True:
            answer = input(question).strip().lower()
            if answer in ("y", "yes"):
                return True
            elif answer in ("n", "no"):
                return False
            else:
                print("Please enter 'y' or 'n'.")
                continue



import click
from pathlib import Path
from getpass import getpass
import json
import psycopg2
import os

CONFIG_PATH = Path.home() / ".acedb" / "config.json"


@click.group()
def cli():
    """CLI entry point for the acedb package."""
    pass


@cli.command()
def login():
    """Login to the database and save configuration."""
    host = click.prompt("Enter the host")
    port = click.prompt("Enter the port", type=int)
    db_name = click.prompt("Enter the database name")
    username = click.prompt("Enter the username")
    password = getpass("Enter the password")

    config = {
        "host": host,
        "port": port,
        "db_name": db_name,
        "username": username,
        "password": password,
    }

    try:
        # Test the connection
        conn = psycopg2.connect(
            host=host,
            port=port,
            dbname=db_name,
            user=username,
            password=password,
            connect_timeout=5,
        )
        conn.close()
        click.echo("Connection successful!")
    except Exception as e:
        click.echo(f"Connection failed: {e}. You may need VPN")
        return

    # Ensure the config directory exists
    CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)

    # Save the configuration to a file
    with open(CONFIG_PATH, "w") as config_file:
        json.dump(config, config_file)


@cli.command()
def logout():
    """Logout from the database."""
    if CONFIG_PATH.exists():
        with open(CONFIG_PATH, "w") as config_file:
            config = json.load(config_file)
            del config["host"]
            del config["port"]
            del config["db_name"]
            del config["username"]
            del config["password"]
            json.dump(config, config_file)

        click.echo("Logged out successfully.")
    else:
        click.echo("No active session found.")


@cli.command()
def check_connection():
    """Check the database connection"""
    if not CONFIG_PATH.exists():
        click.echo("No configuration found. Please login first.")
        return

    with open(CONFIG_PATH, "r") as config_file:
        config = json.load(config_file)

    try:
        conn = psycopg2.connect(
            host=config["host"],
            port=config["port"],
            dbname=config["db_name"],
            user=config["username"],
            password=config["password"],
        )
        conn.close()
        click.echo("Connection successful!")
    except Exception as e:
        click.echo(f"Connection failed: {e}")


@cli.command()
def login_status():
    """Check if the user is logged in."""
    if CONFIG_PATH.exists():
        click.echo("You are logged in.")
    else:
        click.echo("You are not logged in.")


@cli.command()
def list_config():
    """List the current configuration."""
    if CONFIG_PATH.exists():
        with open(CONFIG_PATH, "r") as config_file:
            config = json.load(config_file)
        click.echo(json.dumps(config, indent=4))
    else:
        click.echo("No configuration found.")


@cli.command()
def dbn_login():
    """Adds the Databento API key to env variables file"""

    db_token = click.prompt("Enter your Databento API key")
    db_token = db_token.strip()

    if not db_token:
        click.echo("Databento API key cannot be empty.")
        return

    # Adds API key to the config file
    with open(CONFIG_PATH, "r") as config_file:
        config = json.load(config_file)

    config["dbn_token"] = db_token

    with open(CONFIG_PATH, "w") as config_file:
        json.dump(config, config_file)

    os.environ["DATABENTO_API_KEY"] = db_token
    click.echo("Databento API key added to environment variables.")


@cli.command()
def dbn_logout():
    """Removes the Databento API key from env variables file"""

    with open(CONFIG_PATH, "r") as config_file:
        config = json.load(config_file)
        if "dbn_token" in config:
            del config["db_token"]
            json.dump(config, config_file)
            click.echo("Databento API key removed from environment variables.")
        else:
            click.echo("No Databento API key found.")


if __name__ == "__main__":
    cli()



import psycopg2
import io
from typing import List, Dict, Any, Tuple
import polars as pl
import pandas as pd
from pathlib import Path
from datetime import datetime, timedelta, timezone


TYPE_MAP = {
    "int": "NUMERIC",
    "float": "NUMERIC",
    "string": "VARCHAR(255)",
}


class PostgreDBClient:

    def __init__(self, host, port, db_name, username, password):
        try:
            conn = psycopg2.connect(
                host=host,
                port=port,
                dbname=db_name,
                user=username,
                password=password,
                connect_timeout=5,
            )
            self._cursor = conn.cursor()

        except:
            print("Error connecting to the database. Please check your configuration.")
            raise

        print("Database connection established.")

    def _insert_database(self, dataset: str, schema: str, data: pd.DataFrame) -> None:

        cols = data.columns
        io_buffer = io.StringIO()
        data.to_csv(io_buffer)
        io_buffer.seek(0)

        dataset = self._convert_for_SQL(dataset)
        schema = self._convert_for_SQL(schema)

        copy_query = f' COPY "{dataset}"."{schema}" FROM STDIN WITH CSV HEADER'

        self._cursor.copy_expert(copy_query, io_buffer)
        self._cursor.connection.commit()
        print(f"Data inserted into {dataset}.{schema}.")

    def _ensure_database_schema(
        self, dataset: str, schemas: List[str], col_dict: Dict[str, List[str]]
    ) -> None:
        if not self._check_dataset_in_database(dataset):
            self._create_dataset(dataset)

        for schema in schemas:
            cols = col_dict.get(schema)
            if not self._check_schema_in_database(dataset, schema):
                self._create_schema(dataset, schema, cols)

    def _get_local_range(
        self, dataset: str, schema: str, symbol: str
    ) -> tuple[datetime, datetime]:

        dataset = self._convert_for_SQL(dataset)
        schema = self._convert_for_SQL(schema)
        symbol = self._convert_for_SQL(symbol)

        start_end_query = f'SELECT MIN(ts_event), MAX(ts_event) FROM "{dataset}"."{schema}" WHERE symbol = %s'
        self._cursor.execute(start_end_query, (symbol,))
        start, end = self._cursor.fetchone()

        return start, end

    def _retrieve_data(self, dataset: str, schema: str, symbol: str):
        """
        Get the start and end dates of the schema
        """
        dataset = self._convert_for_SQL(dataset)
        schema = self._convert_for_SQL(schema)
        symbol = self._convert_for_SQL(symbol)

        start_end_query = f'SELECT * FROM "{dataset}"."{schema}" WHERE symbol = %s'
        self._cursor.execute(start_end_query, (symbol,))
        rows = self._cursor.fetchall()
        columns = [col[0] for col in self._cursor.description]
        all_data = pl.DataFrame(rows, schema=columns, orient="row").to_pandas()
        return all_data

    def _find_missing_range(
        self,
        dataset_range: tuple[datetime, datetime],
        database_range: tuple[datetime | None, datetime | None],
        query_range: tuple[datetime | None, datetime | None],
    ) -> list[tuple[datetime, datetime]]:

        ds_start, ds_end = dataset_range
        db_start, db_end = database_range
        q_start, q_end = query_range

        if q_start is None or q_end is None:
            q_start = ds_start
            q_end = ds_end

        if db_start is not None:
            db_start, db_end = self.drop_tz(db_start), self.drop_tz(db_end)
        q_start, q_end = self.drop_tz(q_start), self.drop_tz(q_end)

        # If nothing in DB, download full query interval
        if db_start is None or db_end is None:
            return [(q_start, q_end)]

        missing = []
        if q_start < db_start:
            missing.append((q_start, min(db_start, q_end)))
        if q_end > db_end:
            missing.append((max(db_end, q_start), q_end))

        return missing

    def _check_dataset_in_database(self, dataset: str) -> bool:
        """
        Check if the dataset is in the database
        Databento Dataset is one Schema in DB
        """
        dataset = self._convert_for_SQL(dataset)
        ds_check_query = "SELECT EXISTS (SELECT 1 FROM information_schema.schemata WHERE schema_name = %s) AS dataset_exists"
        self._cursor.execute(ds_check_query, (dataset,))
        exists = self._cursor.fetchone()

        return bool(exists[0])

    def _check_schema_in_database(self, dataset: str, schema: str) -> bool:
        """
        Check if the schema is in the database
        """
        dataset = self._convert_for_SQL(dataset)
        schema = self._convert_for_SQL(schema)

        schma_check_query = "SELECT EXISTS (SELECT 1 FROM information_schema.tables WHERE table_schema = %s AND  table_name = %s ) AS schema_exists"
        self._cursor.execute(schma_check_query, (dataset, schema))
        exists = self._cursor.fetchone()

        return bool(exists[0])

    def _create_dataset(self, dataset: str):
        """
        Create a new dataset in the database
        """
        dataset = self._convert_for_SQL(dataset)

        create_dataset_query = f"CREATE SCHEMA {dataset}"
        self._cursor.execute(create_dataset_query)
        print(f"Dataset {dataset} created.")

    def _create_schema(self, dataset: str, schema: str, cols: List[str]):
        """
        Create a new schema in the database
        """
        dataset = self._convert_for_SQL(dataset)
        schema = self._convert_for_SQL(schema)
        create_schema_query = f'CREATE TABLE IF NOT EXISTS "{dataset}"."{schema}"  '
        col_defs = ",\n    ".join(
            f"{col['name']} {TYPE_MAP.get(col['type'], col['type'])}" for col in cols
        )
        create_schema_query += f"({col_defs})"

        self._cursor.execute(create_schema_query)
        self._cursor.connection.commit()

        print(f"Table {schema} created in Schema {dataset}.")

    def _get_datasets(
        self,
    ):

        query = (
            " SELECT schema_name FROM information_schema.schemata ORDER BY schema_name"
        )
        self._cursor.execute(query)
        rows = self._cursor.fetchall()
        datasets = [row[0] for row in rows]
        return datasets

    def _get_schemas(self, dataset: str) -> List[str]:
        """
        Get the schemas in the database
        """
        dataset = self._convert_for_SQL(dataset)
        query = f"SELECT table_name FROM information_schema.tables WHERE table_schema = '{dataset}'"
        self._cursor.execute(query)
        rows = self._cursor.fetchall()
        schemas = [row[0] for row in rows]
        return schemas

    def _get_symbols(self, dataset: str, schema: str) -> List[str]:
        """
        Get the symbols in the database
        """
        dataset = self._convert_for_SQL(dataset)
        schema = self._convert_for_SQL(schema)
        query = f'SELECT DISTINCT symbol FROM "{dataset}"."{schema}"'
        self._cursor.execute(query)
        rows = self._cursor.fetchall()
        symbols = [row[0] for row in rows]
        return symbols

    def _disconnect(self):
        """
        Disconnect from the database
        """
        self._cursor.close()
        self._cursor.connection.close()
        print("Database connection closed.")

    def _download(self, data_dict: dict, path: str | Path = None, filetype="csv"):
        """
        Download the data from the database
        """
        path = Path(path or ".")

        path.mkdir(parents=True, exist_ok=True)

        ext = filetype.lower()

        for schema, symbols in data_dict.items():
            for symbol, data in symbols.items():

                filepath = (
                    path
                    / f"{schema}_{symbol}_{datetime.now().strftime('%Y%m%d_%H%M%S')}.{ext}"
                )
                writer = getattr(data, f"to_{ext}", None)
                if writer is None:
                    raise ValueError(f"Unsupported file type: {ext}")

                kwargs = {}
                if ext in ("csv", "xls", "xlsx", "html"):
                    kwargs["index"] = False
                elif ext == "parquet":
                    kwargs["compression"] = "gzip"
                elif ext == "json":
                    kwargs["orient"] = "records"

                writer(filepath, **kwargs)
                print(f"Data downloaded to {filepath}")

    @staticmethod
    def _convert_for_SQL(terms: List[str] | str) -> List[str]:
        """
        Convert the terms to a string
        """
        if isinstance(terms, str):
            return terms.replace(".", "_").replace("-", "_")
        else:
            return [term.replace(".", "_").replace("-", "_") for term in terms]

    @staticmethod
    def drop_tz(d: datetime) -> datetime:
        return d.replace(tzinfo=None) if d.tzinfo is not None else d
